import numpy as np
from .array import Array
# from ..utils import value_to_string


class Dataset:
    def __init__(self):
        self._container = {}
        self.meta = {}
        self.shape = None

    def __iter__(self):
        return self._container.__iter__()

    def __len__(self):
        return self._container.__len__()

    def __getitem__(self, key):
        if isinstance(key, str):
            return self._container[key]
        else:
            d = Dataset()
            for name, val in self.items():
                d[name] = val[key]
            d.meta.update(self.meta)
            return d

    def __setitem__(self, key, value):
        shape = value.shape[0]
        if self.shape is None:
            self.shape = shape
        else:
            if self.shape != shape:
                raise RuntimeError(
                    "Size mismatch on element insertion. Item "
                    "shape is {} while container accepts shape {}.".format(
                        shape, self.shape))
        if isinstance(value, Array):
            value.name = key
            value.parent = self
            self._container[key] = value
        else:
            self._container[key] = Array(values=value, name=key, parent=self)

    def __delitem__(self, key):
        return self._container.__delitem__(key)

    def __repr__(self):
        return str(self)

    def __str__(self):
        output = "{}: {}\n".format(self.meta["infile"], self.print_size())
        for key, item in self.items():
            output += str(item) + "\n"
        return output

    def keys(self):
        return self._container.keys()

    def items(self):
        return self._container.items()

    def values(self):
        return self._container.values()

    def set_scale(self, scale):
        for key in ["x", "y", "z", "dx"]:
            if key in self:
                self[key].to(scale)

    def print_size(self):
        total_size = np.sum([item._array.nbytes for item in self.values()])
        multipliers = {"G": 1.0e9, "M": 1.0e6, "K": 1.0e3, "B": 1.0}
        for m, mult in multipliers.items():
            if total_size >= mult:
                size = "{:.2f} {}B".format(total_size / mult, m)
                break
        return size



import numpy as np
from . import utils


class Loader():
    def __init__(self):
        self.variables = {}
        self.offsets = {}
        self.meta = {}
        self.bytes = None
        return

    def allocate_buffers(self, ngridmax, twotondim):
        for item in self.variables.values():
            item["buffer"] = np.zeros([ngridmax, twotondim],
                                      dtype=np.dtype(item["type"]))

    def read_header(self, *args, **kwargs):
        return

    def read_level_header(self, *args, **kwargs):
        return

    def read_domain_header(self, *args, **kwargs):
        return

    def read_cacheline_header(self, *args, **kwargs):
        return

    def read_variables(self, ncache, ind, ilevel, cpuid, info):
        for item in self.variables.values():
            if item["read"]:
                item["buffer"][:ncache, ind] = np.array(
                    utils.read_binary_data(
                        fmt="{}{}".format(ncache, item["type"]),
                        content=self.bytes,
                        offsets=self.offsets)) * item["unit"].magnitude
            else:
                self.offsets[item["type"]] += ncache
                self.offsets["n"] += ncache

    def make_conditions(self, select, ncache):
        conditions = {}
        for key, func in select.items():
            if not isinstance(func, bool):
                if key in self.variables:
                    conditions[key] = func(self.variables[key]["buffer"][:ncache, :])
        return conditions

    def read_footer(self, *args, **kwargs):
        return

    def step_over(self, ncache, twotondim, ndim):
        self.offsets['d'] += ncache * twotondim * len(self.variables)
        self.offsets['n'] += twotondim * len(self.variables)



import numpy as np
from .. import units


def get_unit(string, ud, ul, ut):
    ramses_units = {
        "density": ud * (units.g / (units.cm**3)),
        "velocity": (ul / ut) * (units.cm / units.s),
        "part_velocity": (ul / ut) * (units.cm / units.s),
        "momentum": (ud * ul / ut) * (units.g / (units.cm**2) / units.s),
        "B_": np.sqrt(4.0 * np.pi * ud * (ul / ut)**2) * units.G,
        "part_tracer_b": np.sqrt(4.0 * np.pi * ud * (ul / ut)**2) * units.G,
        "acceleration": (ul / ut**2) * (units.cm / (units.s**2)),
        "thermal_pressure": ud * ((ul / ut)**2) * (units.erg / (units.cm**3)),
        "energy": ud * ((ul / ut)**2) * (units.erg / (units.cm**3)),
        "temperature": 1.0 * units.K
    }
    ramses_units.update(dict.fromkeys(['x', 'y', 'z', 'dx'], (ul * units.cm)))

    if string in ramses_units:
        return ramses_units[string]

    for key in ramses_units:
        if string.startswith(key):
            return ramses_units[key]

    for key in ramses_units:
        if key in string:
            return ramses_units[key]

    return 1.0 * units.dimensionless



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
from ..core import Plot
from .. import units
from .render import render
from ..core.tools import to_bin_centers, finmin, finmax
from .parser import parse_layer, get_norm
from scipy.stats import binned_statistic_2d


def histogram(x,
              y,
              *layers,
              mode=None,
              ax=None,
              logx=False,
              logy=False,
              loglog=False,
              norm=None,
              cbar=True,
              filename=None,
              resolution=256,
              operation="sum",
              title=None,
              xmin=None,
              xmax=None,
              ymin=None,
              ymax=None,
              vmin=None,
              vmax=None,
              **kwargs):
    """
    Plot a 2D histogram with two variables as input.
    """
    if loglog:
        logx = logy = True

    nx = resolution
    ny = resolution

    xvals = x.norm
    yvals = y.norm

    # Define plotting range
    autoxmin = False
    autoxmax = False
    autoymin = False
    autoymax = False

    if xmin is None:
        xmin = finmin(xvals)
        autoxmin = True
    if xmax is None:
        xmax = finmax(xvals)
        autoxmax = True
    if ymin is None:
        ymin = finmin(yvals)
        autoymin = True
    if ymax is None:
        ymax = finmax(yvals)
        autoymax = True

    if logx:
        [xmin, xmax] = np.log10([xmin, xmax])
    if logy:
        [ymin, ymax] = np.log10([ymin, ymax])

    # Protect against empty plots if xmin==xmax or ymin==ymax
    if xmin == xmax:
        if xmin == 0.0:
            xmin = -0.1
            xmax = 0.1
        else:
            xmin = xmin - 0.05 * abs(xmin)
            xmax = xmax + 0.05 * abs(xmax)
    if ymin == ymax:
        if ymin == 0.0:
            ymin = -0.1
            ymax = 0.1
        else:
            ymin = ymin - 0.05 * abs(ymin)
            ymax = ymax + 0.05 * abs(ymax)

    dx = xmax - xmin
    dy = ymax - ymin
    if autoxmin:
        xmin = xmin - 0.05 * dx
    if autoxmax:
        xmax = xmax + 0.05 * dx
    if autoymin:
        ymin = ymin - 0.05 * dy
    if autoymax:
        ymax = ymax + 0.05 * dy

    # Construct some bin edges
    if logx:
        xedges = np.logspace(xmin, xmax, nx + 1)
    else:
        xedges = np.linspace(xmin, xmax, nx + 1)
    if logy:
        yedges = np.logspace(ymin, ymax, ny + 1)
    else:
        yedges = np.linspace(ymin, ymax, ny + 1)

    # In the contour plots, x and y are the centers of the cells, instead of
    # the edges.
    xcenters = to_bin_centers(xedges)
    ycenters = to_bin_centers(yedges)

    to_render = []
    # Buffer for counts
    to_process = [np.ones_like(xvals)]
    operations = ["sum"]

    if layers is not None:
        for layer in layers:
            data, settings, params = parse_layer(layer,
                                                 mode=mode,
                                                 norm=norm,
                                                 vmin=vmin,
                                                 vmax=vmax,
                                                 operation=operation,
                                                 **kwargs)
            to_process.append(data.norm)
            to_render.append({
                "mode": settings["mode"],
                "params": params,
                "unit": data.unit.units,
                "name": data.name
            })
            operations.append(settings["operation"])

    if (operation == "mean") and "sum" in operations:
        counts, _, _ = np.histogram2d(yvals, xvals, bins=(yedges, xedges))

    binned, _, _, _ = binned_statistic_2d(x=yvals,
                                          y=xvals,
                                          values=to_process,
                                          statistic=operation,
                                          bins=[yedges, xedges])

    # Here we assume that dictionary retains order of insertion: counts
    # are the first key
    mask = binned[0] == 0.0
    for ind in range(1, len(to_process)):
        if operations[ind] != operation:
            if operation == "sum":
                with np.errstate(invalid="ignore"):
                    binned[ind] /= binned[0]
            else:
                binned[ind] *= counts
        to_render[ind - 1]["data"] = np.ma.masked_where(mask, binned[ind])

    if len(to_render) == 0:
        to_render.append({
            "data": np.ma.masked_where(mask, binned[0]),
            "mode": mode,
            "name": "counts",
            "params": {
                "norm": get_norm(norm=norm, vmin=vmin, vmax=vmax),
                "vmin": vmin,
                "vmax": vmax
            },
            "unit": units.dimensionless
        })

    figure = render(x=xcenters, y=ycenters, data=to_render, logx=logx, logy=logy)

    figure["ax"].set_xlabel(x.label)
    figure["ax"].set_ylabel(y.label)

    return Plot(x=xcenters,
                y=ycenters,
                layers=to_render,
                fig=figure["fig"],
                ax=figure["ax"])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
import numpy.ma as ma
from pint.quantity import Quantity
from .slice import get_slice_direction
from .render import render
from .parser import parse_layer
from ..core import Plot, Array
from ..core.tools import to_bin_centers, apply_mask
from scipy.stats import binned_statistic_2d
from scipy.ndimage import distance_transform_edt


def plane(*layers,
          direction="z",
          dx=None,
          dy=None,
          fname=None,
          title=None,
          sinks=True,
          plot=True,
          mode=None,
          norm=None,
          vmin=None,
          vmax=None,
          operation="mean",
          origin=None,
          resolution=256,
          ax=None,
          **kwargs):
    """
    Plot a 2D slice through the data domain.
    """

    if isinstance(layers, Array):
        layers = [layers]

    to_process = []
    to_render = []
    operations = []
    for layer in layers:
        data, settings, params = parse_layer(layer,
                                             mode=mode,
                                             norm=norm,
                                             vmin=vmin,
                                             vmax=vmax,
                                             operation=operation,
                                             **kwargs)

        to_process.append(data)
        to_render.append({
            "mode": settings["mode"],
            "params": params,
            "unit": data.unit.units,
            "name": data.name
        })
        operations.append(settings["operation"])

    dataset = to_process[0].parent

    # Set window size
    if dy is None:
        dy = dx
    if dx is not None and not isinstance(dx, Quantity):
        dx *= dataset["xyz"].unit
    if dy is not None and not isinstance(dy, Quantity):
        dy *= dataset["xyz"].unit

    dir_vecs, origin = get_slice_direction(direction=direction,
                                           dataset=dataset,
                                           dx=dx,
                                           dy=dy,
                                           origin=origin)

    # Distance to the plane
    xyz = dataset["xyz"] - origin
    diagonal = dataset["dx"] * np.sqrt(dataset.meta["ndim"]) * 0.5
    dist1 = np.sum(xyz * dir_vecs[0], axis=1)  # / np.linalg.norm(dir_vecs[0][1])

    # Select cells in contact with plane
    cells_in_plane = np.abs(dist1) <= diagonal
    # Project coordinates onto the plane by taking dot product with axes
    # vectors
    select = np.ravel(np.where(cells_in_plane))
    coords = xyz[select]
    datax = np.inner(coords, dir_vecs[1])
    datay = np.inner(coords, dir_vecs[2])
    datadx = 0.5 * dataset["dx"][select]

    # Get limits
    limits = {
        'xmin': np.amin(datax - datadx).values,
        'xmax': np.amax(datax + datadx).values,
        'ymin': np.amin(datay - datadx).values,
        'ymax': np.amax(datay + datadx).values
    }

    # Define slice extent
    if dx is None:
        xmin = limits['xmin']
        xmax = limits['xmax']
        ymin = limits['ymin']
        ymax = limits['ymax']
    else:
        xmin = -0.5 * dx.magnitude
        xmax = xmin + dx.magnitude
        ymin = -0.5 * dy.magnitude
        ymax = ymin + dy.magnitude
        # Limit selection further by using distance from center
        dist2 = xyz - diagonal
        select = np.ravel(
            np.where(
                np.logical_and(
                    cells_in_plane,
                    np.abs(dist2.norm) <=
                    max(dx.magnitude, dy.magnitude) * 0.5 * np.sqrt(2.0))))
        coords = xyz[select]
        datax = np.inner(coords, dir_vecs[1])
        datay = np.inner(coords, dir_vecs[2])

    scalar_layer = []
    to_binning = []
    for ind in range(len(to_process)):
        if to_render[ind]["mode"] in ["vec", "stream"]:
            if to_process[ind].ndim < 3:
                uv = to_process[ind].array[select]
            else:
                uv = np.inner(to_process[ind].array.take(select, axis=0), dir_vecs[1:])
            w = None
            if "color" in to_render[ind]["params"]:
                if isinstance(to_render[ind]["params"]["color"], Array):
                    w = to_render[ind]["params"]["color"].norm
                elif isinstance(to_render[ind]["params"]["color"], np.ndarray):
                    w = to_render[ind]["params"]["color"]
            if w is None:
                w = np.linalg.norm(uv, axis=1)
            else:
                w = w.take(select, axis=0)
            to_binning.append(apply_mask(uv[:, 0]))
            to_binning.append(apply_mask(uv[:, 1]))
            to_binning.append(apply_mask(w))
            scalar_layer.append(False)
        else:
            to_binning.append(apply_mask(to_process[ind].array[select]))
            scalar_layer.append(True)

    # Buffer for counts
    to_binning.append(np.ones_like(to_binning[0]))

    # Construct some bin edges
    xedges = np.linspace(xmin, xmax, resolution + 1)
    yedges = np.linspace(ymin, ymax, resolution + 1)

    # In the contour plots, x and y are the centers of the cells, instead of
    # the edges.
    xcenters = to_bin_centers(xedges)
    ycenters = to_bin_centers(yedges)

    binned, _, _, _ = binned_statistic_2d(x=apply_mask(datay.array),
                                          y=apply_mask(datax.array),
                                          values=to_binning,
                                          statistic="mean",
                                          bins=[yedges, xedges])

    # Use Scipy's distance transform to fill blanks with nearest neighbours
    condition = np.isnan(binned[-1])
    transform = tuple(
        distance_transform_edt(condition, return_distances=False, return_indices=True))

    # Define a mask to mask aread outside of the domain range, which have
    # been filled by the previous transform step
    xx = np.broadcast_to(xcenters, [resolution, resolution])
    yy = np.broadcast_to(ycenters, [resolution, resolution]).T
    mask = np.logical_or.reduce((xx < limits['xmin'], xx > limits['xmax'],
                                 yy < limits['ymin'], yy > limits['ymax']))
    mask_vec = np.broadcast_to(mask.reshape(*mask.shape, 1), mask.shape + (3, ))

    counter = 0
    for ind in range(len(to_render)):
        if scalar_layer[ind]:
            to_render[ind]["data"] = ma.masked_where(mask,
                                                     binned[counter][transform],
                                                     copy=False)
            counter += 1
        else:
            to_render[ind]["data"] = ma.masked_where(
                mask_vec,
                np.array([
                    binned[counter][transform].T, binned[counter + 1][transform].T,
                    binned[counter + 2][transform].T
                ]).T,
                copy=False)
            counter += 3

    to_return = {"x": xcenters, "y": ycenters, "layers": to_render}
    if plot:
        # Render the map
        figure = render(x=xcenters, y=ycenters, data=to_render, ax=ax)
        figure["ax"].set_xlabel(dataset["xyz"].x.label)
        figure["ax"].set_ylabel(dataset["xyz"].y.label)
        if ax is None:
            figure["ax"].set_aspect("equal")
        to_return.update({"fig": figure["fig"], "ax": figure["ax"]})

    return Plot(**to_return)
